////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// jtag.cpp
//
// (c) Lukas Franz Josef (Brain) Schrittwieser
//
// Feb 2006
//
// Published under the terms of the GNU general public license
//
//
// Compiler: Microsoft Visual C++ 6.0
//
// This is the software module for the RS232 to JTAG bridge. The birdge together with this module allows to access
// the registers of the attached JTAG chain. This module is written for the W32 library of Microsoft Visual C++
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// 24. April 2006:
// Some changes in the bridge hardware require a new software module. Note: This software works only with the new
// hardware revision 2 (=com4jtag.lif)
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#include "jtag.h"
#include "usb.h"


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//   I M P L E M E N T A T I O N


// Konstruktor
Cjtag::Cjtag (void) : cmdBufferLen(1024)
{
	cmdIndex = 0;
	usbhandle = 0;
	verbose = 0;

	cmdBuffer = new unsigned char [cmdBufferLen];
}


// Destruktor
Cjtag::~Cjtag (void)
{
	if (cmdBuffer != NULL)
		delete [] cmdBuffer;
	close();
}


// initialize the class
// and opens the usb device
int Cjtag::init () 
{
	usb_init();

	if (usbOpenDevice(&usbhandle, USBASP_SHARED_VID, "www.fischl.de",
		USBASP_SHARED_PID, "USBasp") != 0) 
	{
		/* check if device with old VID/PID is available */
		if (usbOpenDevice(&usbhandle, USBASP_OLD_VID, "www.fischl.de",
			USBASP_OLD_PID, "USBasp") != 0) 
		{

			/* no USBasp found */
			fprintf(stderr, "Cjtag::init: error: could not find USB device \"USBasp\" with vid=0x%x pid=0x%x\n",
  				  USBASP_SHARED_VID, USBASP_SHARED_PID);
			return 0;

		} 
		else
		{
			/* found USBasp with old IDs */
			fprintf(stderr,
				  "Warning: Found USB device \"USBasp\" with "
				"old VID/PID! Please update firmware of USBasp!\n");
		}
	}

	unsigned char cmd[4] = {0};
	unsigned char rply[4] = {0};

	cmd[0] = 6;	// jtag clock:	5: 18mhz div by 32
				//				4: 18mhz div by 16
	cmd[1] = 0;
	usbasp_transmit(1, USBASP_JTAG_CONF, cmd, rply, 1);

	if (rply[0] == 0x01)
		resetTAP();

	return rply[0];	// the programmer returns 1 on success

}

int Cjtag::getState (void)
{
	if (usbhandle == 0)
		return -1;

	execute();

	unsigned char cmd[4] = {0};
	unsigned char rply[4] = {0};

	usbasp_transmit(1, USBASP_JTAG_GET_STATE, cmd, rply, 1);

	return rply[0];	// the programmer returns the tap state
}


void Cjtag::close (void)
{
	unsigned char cmd[4] = {0};
	unsigned char rply[4] = {0};

	execute();

	cmd[0] = 0;	// jtag disable
	cmd[1] = 0;
	
	if (usbhandle)
	{
		usbasp_transmit(1, USBASP_JTAG_CONF, cmd, rply, 1);
		usb_close(usbhandle);
	}
	usbhandle = 0;
}


int Cjtag::runTest (void)
{
	if (usbhandle == 0)
		return -1;
	
	addCmd (JTAG_CM_RUN_TEST, NULL, 0);

	return 0;
}

int Cjtag::accessDR (unsigned char* writeData, unsigned char* readData, int length, bool stayInShiftState, bool readBack)
{
	unsigned char cmd[4] = {0};
	unsigned char rply[4] = {0};
	int shiftedBits = length;

	if (usbhandle == 0)
		return -1;

	//printBuffer(writeData, (length+7)/8);

	while (length>(128*8))
	{
		// send one chunk of 128 bytes
		addCmd (JTAG_CM_SHIFT_DR | 0x08, writeData, 128);
		
		if (readBack)
		{
			// execute the commands (we need the results so the data has to be transferd to the device)
			execute();
			// receive the 128 answer bytes 
			usbasp_transmit(1, USBASP_JTAG_READ_BACK, cmd, readData, 128);
		}
		
		
		writeData += 128;
		readData += 128;
		length -= (128*8);
	}
	// send the last packet
	cmd[0] = length%8;	// number of bits in the last byte 
	if (cmd[0] == 0)
		cmd[0] = 8;

	//printf ("shiftDR: len=%d, bytes=%d, lastBits=%d\n", length, (length+7)/8, cmd[0]);
	

//	getchar();

	addCmd (JTAG_CM_SHIFT_DR | (cmd[0]&0x0F), writeData, (length+7)/8);

	if (readBack)
	{
		// execute the commands (we need the results so the data has to be transferd to the device)
		execute();
		// receive last read back bytes
		usbasp_transmit(1, USBASP_JTAG_READ_BACK, cmd, readData, (length+7)/8);
		// the bits in the last byte are not MSB aligned, correct it here 
		readData[((length+7)/8)-1] <<= (8-cmd[0]);	// cmd[0] is the number of bits in the last byte
	}

	//printf ("state machine is in: %d \n", getState());

	// if we don't stay in shift we move on to update-dr state
	if (!stayInShiftState)
		addCmd (JTAG_CM_UPDATE_DR, NULL, 0);

	//execute();

	return shiftedBits;
}


// This method works like accessDR, but it uses the TAP's instruction register
int Cjtag::accessIR (unsigned char* writeData, unsigned char* readData, int length, bool readBack)
{
	unsigned char cmd[4] = {0};
	unsigned char rply[4] = {0};

	if (usbhandle == 0)
		return -1;


	int shiftedBits = length;

	//printf ("Cjtag::accessIR: length=%d, rb=%d\n", length, (readBack)?1:0);

	
	while (length>128*8)
	{
		// send one chunk of 128 bytes
		// fill the shift command and the data into the buffer, if it becomes full it's sent to the device
		addCmd (JTAG_CM_SHIFT_IR | 0x08, writeData, 128);

		if (readBack)
		{
			execute();
			// receive the 128 answer bytes 
			usbasp_transmit(1, USBASP_JTAG_READ_BACK, cmd, readData, 128);
		}
		writeData += 128;
		readData += 128;
		length -= 128*8;
	}
	// send the last packet
	cmd[0] = length%8;	// number of bits in the last byte 
	if (cmd[0] == 0)	// if the last bit is complete we have 0 here but need 8
		cmd[0] = 8;
	// fill the shift command and the data into the buffer, if it becomes full it's sent to the device
	addCmd (JTAG_CM_SHIFT_IR | (cmd[0]&0x0f), writeData, (length+7)/8);

	if (readBack)
	{
		execute();
		// receive last read back bytes
		usbasp_transmit(1, USBASP_JTAG_READ_BACK, cmd, readData, (length+7)/8);
		// the bits in the last byte are not MSB aligned, correct it here 
		readData[((length+7)/8)-1] <<= 8-cmd[0];	// cmd[0] is the number of bits in the last byte
	}


	// move the tap to update ir state
	addCmd (JTAG_CM_UPDATE_IR, NULL, 0);
	

	//execute();

	return shiftedBits;
}


// This function puts the state machine in the TEST-LOGIC-RESET state
// This resets the TAP
int Cjtag::resetTAP (void)
{
	if (usbhandle == 0)
		return -1;

	addCmd (JTAG_CM_RESET_TAP, NULL, 0);
	return 0;
}


// filles a command into the command buffer
// cmd is the command byte to add
// data is the optional data field, len the number of bytes in data
int Cjtag::addCmd (unsigned char cmd, const unsigned char* data, unsigned char len)
{
	if ((cmdIndex+len) > cmdBufferLen)
		execute();

	if ((cmdIndex+len) > cmdBufferLen)
	{
		fprintf (stderr, "Cjtag: internal error, command is too long for buffer\n");
		return -1;
	}

	// copy the command byte into the buffer
	cmdBuffer[cmdIndex++] = cmd;

	// check if the buffer is full
	if (cmdIndex == cmdBufferLen)
		execute();	// transfer the buffer to the device

	// if we have no data we are done
	if (len == 0)
		return 0;

	// we have data, copy the length byte into the buffer
	cmdBuffer[cmdIndex++] = len;

	// check if the buffer is full
	if (cmdIndex == cmdBufferLen)
		execute();	// transfer the buffer to the device

	// copy the data into the buffer
	while (len > 0)
	{
		// determine the number bytes to copy
		int size = (len < (cmdBufferLen-cmdIndex)) ? len : cmdBufferLen-cmdIndex;
		//printf ("size: %d\n", size);
		// copy the bytes
		memcpy ((void*)(&(cmdBuffer[cmdIndex])), data, size);
		len -= size;
		cmdIndex += size;
		// check if the buffer is full
		if (cmdIndex >= cmdBufferLen)
			execute();	// transfer the buffer to the device	
	
	}
	
	return 0;
}



void Cjtag::printBuffer(unsigned char* buf, int len)
{
	int i;
	for (i=0; i<len; i++)
		printf ("0x%02X ",buf[i]);
	printf(" |  '");
	//for (i=0; i<len; i++)
	//	printf ("%c",(buf[i]>31)?buf[i]:'-');
	printf("'\n");
}



int Cjtag::execute ()
{
	unsigned char cmd[4] = {0};
	
	if (usbhandle == 0)
		return -1;

	if (cmdIndex == 0)
		return 0;

	//printf ("executing with %d bytes\n", cmdIndex);

	//printBuffer(cmdBuffer, cmdIndex);
	//printf ("\n");

	// transfer the whole command buffer to the device
	int retVal = usbasp_transmit(0, USBASP_JTAG_EXECUTE, cmd, cmdBuffer, cmdIndex);
	cmdIndex = 0;	// the buffer is empty
	return retVal;
}


int Cjtag::usbasp_transmit(unsigned char receive, unsigned char functionid,
			   unsigned char send[4], unsigned char * buffer, int buffersize)
{
  int nbytes;
  nbytes = usb_control_msg(usbhandle,
			   USB_TYPE_VENDOR | USB_RECIP_DEVICE | (receive << 7),
			   functionid,
			   (send[1] << 8) | send[0], /* value */
			   (send[3] << 8) | send[2], /* index */
			   (char*)buffer, buffersize,
			   4000);
  if(nbytes < 0){
	fprintf(stderr, "\n%s: error: usbasp_transmit: %s\n", "test1", usb_strerror());
    exit(1);
  }

  return nbytes;
}


int Cjtag::usbOpenDevice(usb_dev_handle **device, int vendor,
			 char *vendorName, int product, char *productName)
{
struct usb_bus      *bus;
struct usb_device   *dev;
usb_dev_handle      *handle = NULL;
int                 errorCode = USB_ERROR_NOTFOUND;
static int          didUsbInit = 0;

    if(!didUsbInit){
        didUsbInit = 1;
        usb_init();
    }
    usb_find_busses();
    usb_find_devices();
    for(bus=usb_get_busses(); bus; bus=bus->next){
        for(dev=bus->devices; dev; dev=dev->next)
		{
            if(dev->descriptor.idVendor == vendor &&
				dev->descriptor.idProduct == product)
			{
                char    string[256];
                int     len;
				/* we need to open the device in order to query strings */
                handle = usb_open(dev);
                if(!handle)	
				{
                    errorCode = USB_ERROR_ACCESS;
                    fprintf(stderr,
					"Warning: cannot open USB device: %s\n",
					usb_strerror());
                    continue;
                }
                if(vendorName == NULL && productName == NULL)
				{
					/* name does not matter */
                    break;
                }
                /* now check whether the names match: */
                len = usb_get_string_simple(handle, dev->descriptor.iManufacturer,
					    string, sizeof(string));
                if(len < 0)
				{
                    errorCode = USB_ERROR_IO;
                    fprintf(stderr, "Warning: cannot query manufacturer for device: %s\n", usb_strerror());
                }
				else
				{
                    errorCode = USB_ERROR_NOTFOUND;
					if (verbose > 1)
						fprintf(stderr, "seen device from vendor ->%s<-\n", string);
                    if(strcmp(string, vendorName) == 0)
					{
                        len = usb_get_string_simple(handle, dev->descriptor.iProduct,string, sizeof(string));
                        if(len < 0)
						{
                            errorCode = USB_ERROR_IO;
                            fprintf(stderr,"Warning: cannot query product for device: %s\n",usb_strerror());
                        }
						else
						{
                            errorCode = USB_ERROR_NOTFOUND;
							if (verbose > 1)
								fprintf(stderr, "seen product ->%s<-\n", string);
                            if(strcmp(string, productName) == 0)
                                break;
                        }
                    }
                }
                usb_close(handle);
                handle = NULL;
            }
        }
        if(handle)
            break;
    }
    if(handle != NULL){
        errorCode = 0;
        *device = handle;
    }
    return errorCode;
}



