/*
	xc9500xl - simple cpld programmer for xilinx xc9536xl and xc9572xl device
	Jtag interface: uJtag (USBasp) 

	date: Feb 2009

	version 1.0

	author: Lukas Schrittwieser

*/

#include <unistd.h>
#include <stdio.h>
#include "jtag.h"

#define ID_XC9536XL		0xC9040690
#define ID_XC9572XL		0xC9020690

// length of the instruction register
#define IREG_LENGTH		8

// isc commands for the cpld
#define ID_CODE			0x7F
#define BLANK_CHECK		0xA7
#define ISC_ENABLE		0x97
#define ISC_DISABLE		0x0F
#define ERASE			0xB7
#define ISC_READ		0x77
#define ISC_PROGRAM		0x57


void waitTck (Cjtag &bridge, int cnt);
unsigned char reverseBits (unsigned char in);
int verifyWord (Cjtag &bridge, int adr, int data);

unsigned long			id;	// device id


void Sleep (int ms)
{
	usleep( 1000 * ms );
}

int blankCheck (Cjtag &bridge)
{
	unsigned char buffer[4] = {BLANK_CHECK, 0x00, 0x00, 0x00};
	unsigned char buffer2[4];

	bridge.accessIR (buffer, NULL, IREG_LENGTH, false);

	buffer[0] = 0xC0;
	buffer[1] = 0;
	buffer[2] = 0;

	bridge.accessDR (buffer, NULL, 18, false, false);
	
	waitTck (bridge, 501);

	bridge.accessDR (buffer, buffer2, 18, false, true);

	//printf ("blank check: 0x%02X%02X%02X\n", buffer2[0], buffer2[1], buffer2[2]&0xC0);
	return (reverseBits(buffer2[0])&0x03);
}

int enableISC (Cjtag &bridge)
{
	unsigned char buffer[4] = {ISC_ENABLE, 0x00, 0x00, 0x00};
	unsigned char buffer2[4];

	bridge.accessIR (buffer, NULL, IREG_LENGTH, false);

	buffer[0] = 0xA8;
	buffer[1] = 0;
	buffer[2] = 0;

	bridge.accessDR (buffer, buffer2, 6, false, true);

	//printf ("enable: 0x%02X\n", buffer2[0]&0xFC);

	waitTck (bridge, 1);
	return 0;
}


int disableISC (Cjtag &bridge)
{
	unsigned char buffer[4] = {ISC_DISABLE, 0x00, 0x00, 0x00};

	bridge.accessIR (buffer, NULL, IREG_LENGTH, false);

	Sleep(100);

	buffer[0] = 0xFF;
	buffer[1] = 0;
	buffer[2] = 0;

	bridge.accessIR (buffer, NULL, IREG_LENGTH, false);

	waitTck (bridge, 1);
	return 0;
}

int erase (Cjtag &bridge)
{
	unsigned char buffer[4] = {ERASE};
	unsigned char buffer2[4];

	bridge.accessIR (buffer, NULL, IREG_LENGTH, false);

	buffer[0] = 0xC0;
	buffer[1] = 0;
	buffer[2] = 0;

	bridge.accessDR (buffer, NULL, 18, false, false);
	
	waitTck (bridge, 1);
	Sleep (500);

	bridge.accessDR (buffer, buffer2, 18, false, true);

	//printf ("erase: 0x%02X%02X%02X\n", buffer2[0], buffer2[1], buffer2[2]&0xC0);

	return 0;
}

int checkId (Cjtag &bridge, unsigned long id)
{	
	unsigned char buffer[4] = {ID_CODE};
	unsigned char buffer2[4];

	bridge.accessIR (buffer, NULL, IREG_LENGTH, false);

	waitTck (bridge, 1);

	bridge.accessDR (buffer, buffer2, 32, false, true);

	printf ("idCode: 0x%02X%02X%02X%02X ", buffer2[0], buffer2[1], buffer2[2], buffer2[3]&0xF0);

	if (((buffer2[0]<<24)|(buffer2[1]<<16)|(buffer2[2]<<8)|(buffer2[3]&0xF0)) == id)
	{
		printf("passed\n");
		return 1;
	}
	else
	{
		printf ("failed, wrong / unsupported device\n");
		return 0;
	}
}

int readLine (Cjtag &bridge, int adr)
{
	unsigned char buffer[5];
	buffer[0] = ISC_READ;
	unsigned char buffer2[5];

	bridge.accessIR (buffer, NULL, IREG_LENGTH, false);

	buffer[0] = 0xFF;
	buffer[1] = 0xFF;
	buffer[2] = reverseBits((adr<<2)&0xFC | 0x03);
	buffer[3] = reverseBits((adr>>6)&0xFF);
	buffer[4] = reverseBits((adr>>14)&0x03);

	bridge.accessDR (buffer, buffer2, 34, false, true);

	waitTck (bridge, 1);

	buffer[0] = 0xFF;
	buffer[1] = 0xFF;
	buffer[2] = reverseBits((adr<<2)&0xFC | 0x03);
	buffer[3] = reverseBits((adr>>6)&0xFF);
	buffer[4] = reverseBits((adr>>14)&0x03);

	bridge.accessDR (buffer, buffer2, 34, false, true);
	printf ("readLine (0x%04X): 0x%02X%02X%02X%02X%02X\n", adr, buffer2[0], buffer2[1], buffer2[2], buffer2[3], buffer2[4]&0xC0);
	buffer2[0] = reverseBits (buffer2[0]);
	buffer2[1] = reverseBits (buffer2[1]);
	buffer2[2] = reverseBits (buffer2[2]);
	return ((((int)buffer2[0])>>2) | (((int)buffer2[1])<<6)  | (((int)buffer2[2])<<14)&0x03);

}


// verifies if location adr contains data
// returns 1 on success and 0 on fail
int verifyWord (Cjtag &bridge, int adr, int data)
{
	unsigned char buffer[10];
	buffer[0] = ISC_READ;
	unsigned char buffer2[10];
	unsigned int len, i;

	bridge.accessIR (buffer, NULL, IREG_LENGTH, false);

	if (id == ID_XC9536XL)
	{
		len = 34;
		buffer[0] = reverseBits((data<<2) | 0x03);
		buffer[1] = reverseBits((data>>6)&0xFF);
		buffer[2] = reverseBits((adr<<2)&0xFC | ((data>>14)&0x03));
		buffer[3] = reverseBits((adr>>6)&0xFF);
		buffer[4] = reverseBits((adr>>14)&0x03);
	}
	else
	{
		len = 50;
		buffer[0] = reverseBits((data<<2) | 0x03);
		buffer[1] = reverseBits((data>>6)&0xFF);
		buffer[2] = reverseBits((data>>14)&0xFF);
		buffer[3] = reverseBits((data>>22)&0xFF);
		buffer[4] = reverseBits((adr<<2)&0xFC | ((data>>30)&0x03));
		buffer[5] = reverseBits((adr>>6)&0xFF);
		buffer[6] = reverseBits((adr>>14)&0x03);

	}

	bridge.accessDR (buffer, buffer2, len, false, false);

	waitTck (bridge, 1);

	bridge.accessDR (buffer, buffer2, len, false, true);
	
	for (i=0; i<(len+7)/8; i++)
		buffer2[i] = reverseBits (buffer2[i]);
	
	
	//printf ("verfiyWord (0x%04X): 0x%02X%02X%02X%02X%02X\n", adr, buffer2[4]&0x03, buffer2[3], buffer2[2], buffer2[1], buffer2[0]);
	
	// check against the data we've got
	if (id == ID_XC9536XL)
	{
		if ( ((buffer2[0]>>2)==(data&0x3F)) && (buffer2[1]==((data>>6)&0xFF)) && ((buffer2[2]&0x03)==((data>>14)&0x03)) )
			return 1;
		else
			return 0;
	}
	else
	{
		if ( (((buffer2[0]>>2)&0x3F)==(data&0x3F)) && (buffer2[1]==((data>>6)&0xFF)) && (buffer2[2]==((data>>14)&0xFF)) 
		  && (buffer2[3]==((data>>22)&0xFF)) && ((buffer2[4]&0x03)==((data>>30)&0x03)) )
			return 1;
		else
			return 0;

	}
}

/* sends one 16-bit data word to address adr */
int writeWord (Cjtag &bridge, unsigned int adr, unsigned int data, unsigned int firstBits, bool readBack)
{
	unsigned char buffer[10];
	buffer[0] = ISC_PROGRAM;
	unsigned char buffer2[10];
	unsigned int len = 0;
	
	bridge.accessIR (buffer, NULL, IREG_LENGTH, false);

	if (id == ID_XC9536XL)
	{
		len = 34;
		buffer[0] = reverseBits(((data<<2)&0xFC) | (firstBits&0x03));
		buffer[1] = reverseBits((data>>6)&0xFF);
		buffer[2] = reverseBits((adr<<2)&0xFC | ((data>>14)&0x03));
		buffer[3] = reverseBits((adr>>6)&0xFF);
		buffer[4] = reverseBits((adr>>14)&0x03);
	}
	else
	{
		len = 50;
		buffer[0] = reverseBits(((data<<2)&0xFC) | (firstBits&0x03));
		buffer[1] = reverseBits((data>>6)&0xFF);
		buffer[2] = reverseBits((data>>14)&0xFF);
		buffer[3] = reverseBits((data>>22)&0xFF);
		buffer[4] = reverseBits((adr<<2)&0xFC | ((data>>30)&0x03));
		buffer[5] = reverseBits((adr>>6)&0xFF);
		buffer[6] = reverseBits((adr>>14)&0x03);
	}

	bridge.accessDR (buffer, buffer2, len, false, false);
	
		
	waitTck (bridge, 2);

	if (readBack)
	{
		Sleep (50);
		bridge.accessDR (buffer, buffer2, len, false, true);
		//printf ("'0x%02X' ", reverseBits(buffer2[0])&0x03);
	}

	return (reverseBits(buffer2[0])&0x03);

}

unsigned char reverseBits (unsigned char in)
{
	unsigned char ret = 0;
	for (int i=0; i<8; i++)
		ret |= (in&(1<<i)) ? (1<<(7-i)) : 0;
	return ret;
}

/* convert an hex ascii char to int */
/* returns -1 for illegal chars */
int convCh (char ch)
{
	if ((ch >= '0') && (ch <= '9'))
		return (ch - '0');
	if ((ch >= 'a') && (ch <= 'f'))
		return (ch - 'a' + 10);
	if ((ch >= 'A') && (ch <= 'F'))
		return (ch - 'A' + 10);

	return -1;
}

/* reads the isc file */
/* the cpld device code is stored in id */
int readIscFile (const char* fn, unsigned int* fb)
{
	char buf[100];
	int n, ind;
	unsigned int  val;	// used to assemble on program value
	int f_ind = 0;	// index for configuration buffer
	int len;
	FILE* f;

	if ((fn==NULL) || (fb==NULL))
		return -1;

	f = fopen(fn, "r");
	if (f==NULL)
	{
		printf ("can't open file\n");
		return -1;
	}

	// check if the file is for the correct device 

	// search for the keyword Entity
	memset (buf, 0, 10);
	do
	{
		if (feof(f))
		{
			printf ("can't find device type, file seems corrupted\n");
			return -1;
		}
		// move the old chars one position foreward
		for (n=4;n>=0;n--)
			buf[n+1] = buf[n];
		// get the next char from the file 
		buf[0] = fgetc(f);
	} while (strcmp (buf, "ytitnE") != 0);	// loop until we find the keyword

	// remove wihte spaces
	do
		buf[0] = fgetc(f);
	while ((buf[0]==' ') || (buf[0]=='\t'));

	// we got the first char of the cpld chip code
	fgets (buf+1, 8, f);	// the chip code is 8 chars long
	//buf[8] = 0;

	if (strcmp (buf, "xc9536xl") == 0)
	{
		id = ID_XC9536XL;
	}
	else if (strcmp (buf, "xc9572xl") == 0)
	{
		id = ID_XC9572XL;
	}
	else
	{
		printf ("Error: isc file specifies an unsupported CPLD, only xc9536xl and xc9572xl are supported\n");
		printf ("Type found: '%s'", buf);
		return -1;
	}


	// search for the keyword 'array'
	memset (buf, 0, 10);
	do
	{
		if (feof(f))
		{
			printf ("can't find data array, file seems corrupted\n");
			return -1;
		}
		// move the old chars one position foreward
		for (n=3;n>=0;n--)
			buf[n+1] = buf[n];
		// get the next char from the file 
		buf[0] = fgetc(f);
	} while (strcmp (buf, "yarra") != 0);	// loop until we find the key word

	// search for the keyword 'Initialize'
	memset (buf, 0, 20);
	do
	{
		if (feof(f))
		{
			printf ("can't find keyword Initialize, file seems corrupted\n");
			return -1;
		}
		// move the old chars one position foreward
		for (n=8;n>=0;n--)
			buf[n+1] = buf[n];
		// get the next char from the file 
		buf[0] = fgetc(f);
	} while (strcmp (buf, "ezilaitinI") != 0);	// loop until we find the key word

	// search for an '(' char
	do
	{
		if (feof(f))
		{
			printf ("can't find data array, file seems corrupted\n");
			return -1;
		}
		buf[0] = fgetc(f);
	} while (buf[0] != '(');

	while ((!feof(f)) && (f_ind<1620))
	{
		memset (buf, 0, 100);
		fgets(buf, 99, f);
		ind = 0;	// reset index
		len = strlen(buf); // read one line
		// parse the line
		while (ind < len)
		{
			while ((n=convCh(buf[ind++])) == -1)
				if ((buf[ind-1] == 0) || (ind>=len))	// check for end of line
					break; // get next line
			if ((buf[ind-1]==0) || (ind>=len))
				break;
			val = n&0x000F;
			val <<= 4;
			n = convCh(buf[ind++]);
			if (n == -1)
			/*{
				printf ("file seems corrupted, please check found '%c'\n", buf[ind-1]);
				return -1;
			}*/
				continue;
			val |= n&0x000F;
			val <<= 4;
			n = convCh(buf[ind++]);
			if (n == -1)
			/*{
				printf ("file seems corrupted, please check found '%c'\n", buf[ind-1]);
				return -1;
			}*/
				continue;
			val |= n&0x000F;
			val <<= 4;
			n = convCh(buf[ind++]);
			if (n == -1)
			/*{
				printf ("file seems corrupted, please check found '%c'\n", buf[ind-1]);
				return -1;
			}*/
				continue;
			val |= n&0x000F;
			if (id == ID_XC9536XL)
			{	// the xc9536 device has 16 bit words 
				
				// store the received value
				fb[f_ind++] = val;
				continue;
			}
			// the xc9572xl has 32 bit words
			val <<= 4;
			n = convCh(buf[ind++]);
			if (n == -1)
				continue;
			val |= n&0x000F;

			val <<= 4;
			n = convCh(buf[ind++]);
			if (n == -1)
				continue;
			val |= n&0x000F;
			
			val <<= 4;
			n = convCh(buf[ind++]);
			if (n == -1)
				continue;
			val |= n&0x000F;
			
			val <<= 4;
			n = convCh(buf[ind++]);
			if (n == -1)
				continue;
			val |= n&0x000F;

			fb[f_ind++] = val;
		}
	}	

	return f_ind;
}

int program (Cjtag &bridge, unsigned int *fileBuffer)
{
	int i=0, j=-12, k, p=0;

	printf ("programming:\n  0%%");
	while (i<1620)
	{
		j+=12;
		writeWord(bridge, j++, fileBuffer[i++], 1, false);
		writeWord(bridge, j++, fileBuffer[i++], 1, false);
		writeWord(bridge, j++, fileBuffer[i++], 1, false);
		writeWord(bridge, j++, fileBuffer[i++], 1, false);
		writeWord(bridge, j++, fileBuffer[i++], 1, false);
		j+=3;
		writeWord(bridge, j++, fileBuffer[i++], 1, false);
		writeWord(bridge, j++, fileBuffer[i++], 1, false);
		writeWord(bridge, j++, fileBuffer[i++], 1, false);
		writeWord(bridge, j++, fileBuffer[i++], 1, false);
		writeWord(bridge, j++, fileBuffer[i++], 1, false);
		j+=3;
		writeWord(bridge, j++, fileBuffer[i++], 1, false);
		writeWord(bridge, j++, fileBuffer[i++], 1, false);
		writeWord(bridge, j++, fileBuffer[i++], 1, false);
		writeWord(bridge, j++, fileBuffer[i++], 1, false);

		writeWord(bridge, j, fileBuffer[i], 3, false);
		bridge.execute();
		Sleep(60);
		writeWord(bridge, j, fileBuffer[i], 0, false);
		bridge.execute();
		Sleep(60);
		k=0;
		while ((writeWord(bridge, j, fileBuffer[i], 0, true) != 1) && (k<32))
			k++;
		// show a progress count 
		if ((i*100/1620) > p)
		{
			p=i*100/1620;
			printf ("\b\b\b\b%3d%%",p);
			fflush(stdout);
		}
		if (k >= 32)
			i=1620;

		i++;
	}

	//printf ("\n");
	
	if (k >= 32)
		return -1;

	printf ("\b\b\b\b100%\n");
	fflush (stdout);

	return 0;
}



int verify (Cjtag &bridge, unsigned int* fileBuffer)
{
	int i=0, j=-11, p=0;

	printf ("verifying:\n  0%%");
	while (i<1620)
	{
		j+=11;
		if (!verifyWord(bridge, j++, fileBuffer[i++])) break;
		if (!verifyWord(bridge, j++, fileBuffer[i++])) break;
		if (!verifyWord(bridge, j++, fileBuffer[i++])) break;
		if (!verifyWord(bridge, j++, fileBuffer[i++])) break;
		if (!verifyWord(bridge, j++, fileBuffer[i++])) break;
		j += 3;
		if (!verifyWord(bridge, j++, fileBuffer[i++])) break;
		if (!verifyWord(bridge, j++, fileBuffer[i++])) break;
		if (!verifyWord(bridge, j++, fileBuffer[i++])) break;
		if (!verifyWord(bridge, j++, fileBuffer[i++])) break;
		if (!verifyWord(bridge, j++, fileBuffer[i++])) break;
		j += 3;
		if (!verifyWord(bridge, j++, fileBuffer[i++])) break;
		if (!verifyWord(bridge, j++, fileBuffer[i++])) break;
		if (!verifyWord(bridge, j++, fileBuffer[i++])) break;
		if (!verifyWord(bridge, j++, fileBuffer[i++])) break;
		if (!verifyWord(bridge, j++, fileBuffer[i++])) break;
		
		// show a progress count 
		if ((i*100/1620) > p)
		{
			p=i*100/1620;
			printf ("\b\b\b\b%3d%%",p);
			fflush (stdout);
		}
	}

	if (i<1620)
	{
		printf ("i: %d\n", i);
		return -1;
	}

	printf ("\n");
	fflush (stdout);

	return 0;
}

/*
int verify (Cjtag &bridge, unsigned short* fileBuffer)
{
	int i=0, j=-11;

	while (i<1620)
	{
		j+=11;
		if (readLine(bridge, j++) != fileBuffer[i++]) break;
		if (readLine(bridge, j++) != fileBuffer[i++]) break;
		if (readLine(bridge, j++) != fileBuffer[i++]) break;
		if (readLine(bridge, j++) != fileBuffer[i++]) break;
		if (readLine(bridge, j++) != fileBuffer[i++]) break;
		j += 3;
		if (readLine(bridge, j++) != fileBuffer[i++]) break;
		if (readLine(bridge, j++) != fileBuffer[i++]) break;
		if (readLine(bridge, j++) != fileBuffer[i++]) break;
		if (readLine(bridge, j++) != fileBuffer[i++]) break;
		if (readLine(bridge, j++) != fileBuffer[i++]) break;
		j += 3;
		if (readLine(bridge, j++) != fileBuffer[i++]) break;
		if (readLine(bridge, j++) != fileBuffer[i++]) break;
		if (readLine(bridge, j++) != fileBuffer[i++]) break;
		if (readLine(bridge, j++) != fileBuffer[i++]) break;
		if (readLine(bridge, j++) != fileBuffer[i++]) break;
	}

	if (i<1620)
		return -1;
	return 0;
}
*/

int loopBackTest (Cjtag &bridge)
{
	int i;
	unsigned char buffer[5];
	unsigned char buffer2[5];

	printf ("performing jtag-bypass loop test\n");
	buffer[0] = 0xFF;
	if (bridge.accessIR(buffer, buffer2, IREG_LENGTH, false)==0)
		return 1;

	//srand (GetTickCount());
	srand (0);

	for (i=0; i<5; i++)
		buffer[i] = (unsigned char)rand();
	
	// shift the test data
	bridge.accessDR(buffer, buffer2, 36, false, true);
	for (i=0; i<5; i++)
	{
		//j = (buffer2[i]&0x80) ? 1 : 0;
		buffer2[i] <<= 1;
		buffer2[i] |= (buffer2[i+1]&0x80) ? 1: 0;
	}
	printf ("buffer:  0x%02X%02X%02X%02X%02X\n", buffer[0], buffer[1], buffer[2], buffer[3], buffer[4]&0xF0);
	printf ("buffer2: 0x%02X%02X%02X%02X%02X\n", buffer2[0], buffer2[1], buffer2[2], buffer2[3], buffer2[4]&0xF0);

	getchar();

	buffer[0] = 0xFF;
	//if (bridge.accessIR(buffer, buffer2, IREG_LENGTH, false)==0)
	//	return 1;

	
	for (i=0; i<5; i++)
		buffer[i] = (unsigned char)rand();
	
	// shift the test data
	bridge.accessDR(buffer, buffer2, 36, false, true);
	for (i=0; i<5; i++)
	{
		//j = (buffer2[i]&0x80) ? 1 : 0;
		buffer2[i] <<= 1;
		buffer2[i] |= (buffer2[i+1]&0x80) ? 1: 0;
	}
	printf ("buffer:  0x%02X%02X%02X%02X%02X\n", buffer[0], buffer[1], buffer[2], buffer[3], buffer[4]&0xF0);
	printf ("buffer2: 0x%02X%02X%02X%02X%02X\n", buffer2[0], buffer2[1], buffer2[2], buffer2[3], buffer2[4]&0xF0);
	

	bridge.execute();
	
	return 0;
}


int main (int argc, char* argv[])
{
	Cjtag			bridge;
	//unsigned long	id;
	int				i;
	//DWORD			time;
	char*			filename;
	unsigned int	fileBuffer[2048];

	// init the bridge
	if (bridge.init()==0)
	{
		fprintf (stderr, "Can't init the bridge hardware, abort\n");
		exit(1);
	}

	
	// check for a filename
	if (argc != 2)
	{
		printf ("Usage: %s [filename]\n", argv[0]);
		return 0;
	}
	else
		filename = argv[1];
	
	printf ("supported devices: xc9536xl, xc9572xl\n");
	printf ("downloading file '%s'\n", filename);

	if (i=readIscFile(filename, fileBuffer) != 1620)
	{
		printf ("isc file error: %d\n", i);
		exit (1);
	}

	

	//loopBackTest (bridge);
	
	if (!checkId(bridge, id))
		goto stop;

	
	enableISC(bridge);
	
	erase(bridge);
	if (blankCheck(bridge) != 0x01)
		printf ("blank check: failed\n");
	else
		printf ("blank check: passed\n");

	
	program (bridge, fileBuffer);

	if (verify (bridge, fileBuffer) != 0)
		printf ("verfiy failed\n");
	else
		printf ("verify passed\n");

	

	disableISC(bridge);


stop:
	bridge.resetTAP();

	// free resources
	bridge.close();
	

	return 0;
}


void waitTck (Cjtag &bridge, int cnt)
{
	for (int i=0; i<cnt+1; i++)
		bridge.runTest();
	bridge.execute();
}


